#![cfg_attr(not(debug_assertions), windows_subsystem = "windows")]

use bracket_lib::prelude::*;
use sidebar::draw_sidebar;
use sound::{Sound, SoundEffect, SoundEffectSamples, SoundSystem};
use specs::prelude::*;
use specs_derive::Component;
use std::{
    cmp::{max, min},
    time::{Duration, Instant},
};
use vga_color::*;

use crate::sound::SoundType;

mod sidebar;
mod sound;
pub mod vga_color;

#[derive(Default)]
struct LevelNumber(u32);

struct SoundEffects {
    step_sound: SoundEffectSamples,
    pickup_sound: SoundEffectSamples,
}

#[derive(Component)]
struct Position {
    x: i32,
    y: i32,
}

#[derive(Component)]
struct Renderable {
    glyph: FontCharType,
    fg: RGB,
    bg: RGB,
}

#[derive(Component)]
struct LeftMover {}

#[derive(Component, Debug)]
struct Player {
    last_moved: Instant,
    score: u32,
    gems: u32,
    whips: u32,
    teleports: u32,
    keys: u32,
}

struct State {
    ecs: World,
    sound_system: SoundSystem,
}

fn try_move_player(delta_x: i32, delta_y: i32, ecs: &mut World, sound_system: &mut SoundSystem) {
    let mut positions = ecs.write_storage::<Position>();
    let mut players = ecs.write_storage::<Player>();

    let mut stepped = false;

    for (player, pos) in (&mut players, &mut positions).join() {
        let now = Instant::now();
        if now - player.last_moved > Duration::from_secs_f32(1. / 7.5) {
            pos.x = min(65, max(0, pos.x + delta_x));
            pos.y = min(24, max(0, pos.y + delta_y));
            stepped = true;
            player.last_moved = now;
        }
    }

    if stepped {
        let sound_effects = ecs.fetch::<SoundEffects>();
        let step_sound = sound_effects.step_sound.clone();
        let _ = sound_system.play_sound(step_sound);
    }
}

fn player_input(gs: &mut State, ctx: &mut BTerm) {
    // Player movement
    match ctx.key {
        None => {} // Nothing happened
        Some(key) => match key {
            VirtualKeyCode::Left | VirtualKeyCode::J => {
                try_move_player(-1, 0, &mut gs.ecs, &mut gs.sound_system)
            }
            VirtualKeyCode::U => try_move_player(-1, -1, &mut gs.ecs, &mut gs.sound_system),
            VirtualKeyCode::Up | VirtualKeyCode::I => {
                try_move_player(0, -1, &mut gs.ecs, &mut gs.sound_system)
            }
            VirtualKeyCode::O => try_move_player(1, -1, &mut gs.ecs, &mut gs.sound_system),
            VirtualKeyCode::Right | VirtualKeyCode::K => {
                try_move_player(1, 0, &mut gs.ecs, &mut gs.sound_system)
            }
            VirtualKeyCode::Comma => try_move_player(1, 1, &mut gs.ecs, &mut gs.sound_system),
            VirtualKeyCode::Down | VirtualKeyCode::M => {
                try_move_player(0, 1, &mut gs.ecs, &mut gs.sound_system)
            }
            VirtualKeyCode::N => try_move_player(-1, 1, &mut gs.ecs, &mut gs.sound_system),
            VirtualKeyCode::S => {
                let sound_effects = gs.ecs.fetch::<SoundEffects>();
                let step_sound = sound_effects.pickup_sound.clone();
                let _ = gs.sound_system.play_sound(step_sound);
            }
            _ => {}
        },
    }
}

impl GameState for State {
    fn tick(&mut self, ctx: &mut BTerm) {
        ctx.cls();

        player_input(self, ctx);
        self.run_systems();

        let positions = self.ecs.read_storage::<Position>();
        let renderables = self.ecs.read_storage::<Renderable>();

        for (pos, render) in (&positions, &renderables).join() {
            ctx.set(pos.x, pos.y, render.fg, render.bg, render.glyph);
        }

        draw_sidebar(&self.ecs, ctx);
    }
}

struct LeftWalker {}

impl<'a> System<'a> for LeftWalker {
    type SystemData = (ReadStorage<'a, LeftMover>, WriteStorage<'a, Position>);

    fn run(&mut self, (lefty, mut pos): Self::SystemData) {
        for (_lefty, pos) in (&lefty, &mut pos).join() {
            pos.x -= 1;
            if pos.x < 0 {
                pos.x = 65;
            }
        }
    }
}

impl State {
    fn run_systems(&mut self) {
        let mut lw = LeftWalker {};
        lw.run_now(&self.ecs);
        self.ecs.maintain();
    }
}

fn main() -> BError {
    let context = BTermBuilder::simple(80, 25)?
        .with_fps_cap(60.0)
        .with_title("Kroz")
        .with_tile_dimensions(8, 16)
        .build()?;
    let ss = SoundSystem::new();
    let step_sound = ss.render_sound_effect(SoundEffect {
        sounds: vec![
            Sound {
                sound_type: SoundType::Noise(350, 900),
                duration: Duration::from_millis(6),
            },
            Sound {
                sound_type: SoundType::Silence,
                duration: Duration::from_millis(120),
            },
            Sound {
                sound_type: SoundType::Noise(150, 200),
                duration: Duration::from_millis(6),
            },
        ],
    });

    let pickup_sound = ss.render_sound_effect(SoundEffect {
        sounds: vec![
            Sound {
                sound_type: SoundType::Noise(350, 900),
                duration: Duration::from_millis(6),
            },
            Sound {
                sound_type: SoundType::Silence,
                duration: Duration::from_millis(120),
            },
            Sound {
                sound_type: SoundType::Noise(1000, 2000),
                duration: Duration::from_millis(20),
            },
        ],
    });

    let mut gs = State {
        ecs: World::new(),
        // sound_sender: tx,
        sound_system: ss,
        //step_sound,
    };

    gs.ecs.insert(LevelNumber(0));
    gs.ecs.insert(SoundEffects {
        step_sound,
        pickup_sound,
    });

    gs.ecs.register::<Position>();
    gs.ecs.register::<Renderable>();
    gs.ecs.register::<LeftMover>();
    gs.ecs.register::<Player>();

    gs.ecs
        .create_entity()
        .with(Position { x: 40, y: 24 })
        .with(Renderable {
            glyph: to_cp437('☻'),
            fg: RGB::named(VGA_YELLOW_BRIGHT),
            bg: RGB::named(VGA_BLACK),
        })
        .with(Player {
            last_moved: Instant::now(),
            score: 1290,
            gems: 14,
            whips: 7,
            teleports: 0,
            keys: 0,
        })
        .build();

    for i in 0..10 {
        gs.ecs
            .create_entity()
            .with(Position { x: i * 7, y: 20 })
            .with(Renderable {
                glyph: to_cp437('Ä'),
                fg: RGB::named(VGA_RED_BRIGHT),
                bg: RGB::named(VGA_BLACK),
            })
            .with(LeftMover {})
            .build();
    }

    // let descent_sounds: Vec<Sound> = (20..100)
    //     .rev()
    //     .flat_map(|x| {
    //         (1..10).rev().flat_map(move |y| {
    //             vec![Sound {
    //                 sound_type: SoundType::Tone(x * y * y),
    //                 duration: Duration::from_millis((y as f64 / 1.5) as u64),
    //             }]
    //         })
    //     })
    //     .collect();

    // let descent_effect = gs.sound_system.render_sound_effect(SoundEffect {
    //     sounds: descent_sounds,
    // });

    // let _ = gs.sound_system.play_sound(descent_effect);

    let start_sounds: Vec<Sound> = (30..400)
        .step_by(8)
        .map(|x| Sound {
            sound_type: SoundType::Tone(x),
            duration: Duration::from_millis(24),
        })
        .collect();

    let swoop_effect = gs.sound_system.render_sound_effect(SoundEffect {
        sounds: start_sounds,
    });

    let _ = gs.sound_system.play_sound(swoop_effect);

    // let effect = gs.sound_system.render_sound_effect(SoundEffect {
    //     sounds: vec![Sound {
    //         sound_type: SoundType::Tone(3500),
    //         duration: Duration::from_millis(4000),
    //     }],
    // });

    // let _ = gs.sound_system.play_sound(effect);

    main_loop(context, gs)
}
